package mpredict;

import java.io.IOException;

import java.text.DecimalFormat;

import java.util.Collection;

import java.util.Collections;

import static mpredict.Options.*;

import mpredict.Predict.CutoffBasis;

/**
 * This class handles basic command line processing and calling of main objects.
 * MPredict uses predict class which will predict frequent combinations of
 * putative binding sites when given proper parameters.
 * Requires sequences and PFMs for scoring and locating potential motif sites.
 * Erik Ladewig 8/1/2010
 */

public class MPredict {
    public MPredict() {
        super();
    }

    public static void main(String[] args) {
        String userdir          = System.getProperty("user.dir");
        String inFileMatrix     = null;
        String inFileCutoff     = null;
        String inFileSequences  = null;
        String inFileRandom     = null;
        String outFileDat       = null;
        String outFileFim       = null;
        String outFileSig       = null;

        float pseudo        = 0.375f;
        long time           = System.currentTimeMillis();
        
        int p = 0,
            c = 0,
            s = 0,
            r = 0;
        float percentage = .60f;
        int randSize = 0;
        boolean formatLong = false;
        boolean useRandom = false;
        CutoffBasis basis = CutoffBasis.PWM;

        try{
          Predict pre = new Predict();
          Options opt = new Options(args, 2);

          opt.getSet().addOption("c", Separator.BLANK, Multiplicity.ZERO_OR_ONE);
          opt.getSet().addOption("r", Separator.BLANK, Multiplicity.ZERO_OR_ONE);
          opt.getSet().addOption("l", Multiplicity.ZERO_OR_ONE);
          opt.getSet().addOption("per", Separator.BLANK, Multiplicity.ZERO_OR_ONE);
          opt.getSet().addOption("rsize", Separator.BLANK, Multiplicity.ZERO_OR_ONE);
          opt.getSet().addOption("b", Separator.BLANK, Multiplicity.ZERO_OR_ONE);

          if(!opt.check()) {
            doUserExit(opt.getCheckErrors());
          }

        /**Required Parameters**/
          inFileMatrix = opt.getSet().getData().get(0);
          p = pre.loadPWMs(inFileMatrix, pseudo);
          inFileSequences = opt.getSet().getData().get(1);
          s = pre.loadSequences(inFileSequences, Sequence.SequenceType.STANDARD);
          outFileDat = inFileSequences+".dat";
          outFileFim = inFileSequences+".fim";          
          outFileSig = inFileSequences+".sig";

        /**Optional Parameters**/
          if (opt.getSet().isSet("c") && opt.getSet().isSet("r")) {
            doUserExit("You may not specify use of both a cutoff file (-c) and random cutoff file (-r).\n" +
              "Please choose one of the other.");
          }
          if (opt.getSet().isSet("c")) {
            inFileCutoff = opt.getSet().getOption("c").getResultValue(0);
            c = pre.loadMotifCutOffs(inFileCutoff);
          }
          else if (opt.getSet().isSet("r")) {
            if(opt.getSet().isSet("b")){
               String x = opt.getSet().getOption("b").getResultValue(0).toLowerCase();
               if(x.equals("pwm") || x.equals("p"))
                 basis = CutoffBasis.PWM;
               else if(x.equals("sequence") || x.equals("s"))
                 basis = CutoffBasis.SEQUENCE;
               else
                 doUserExit("Please enter a valid cutoff: (p)wm or (s)equence\n");
            }
            inFileRandom = opt.getSet().getOption("r").getResultValue(0);
            if(inFileRandom == null)
              inFileRandom = inFileSequences;
            useRandom = true;
          }
          else{
            doUserExit("You must use a defined cutoff or generate one with the random option.");
          }
          
          if(opt.getSet().isSet("l")) {
            formatLong = true;
          }
          if(opt.getSet().isSet("per")) {
            percentage = Float.parseFloat( opt.getSet().getOption("per").getResultValue(0) );
          }          
          pre.setPercentCut(percentage);

        /** Brief Output displays relevant information **/
          DecimalFormat df = new DecimalFormat("00.00%");
          System.out.println("Read in "+p+" PWMS, "+c+" cutoffs, "+s+" Sequences.");
          System.out.println("Using sequence file "+inFileSequences+" and "+df.format(percentage)+" cutoff");                      

        /** Run the program **/
          if(useRandom){
            if(opt.getSet().isSet("rsize"))
              randSize = Integer.parseInt(opt.getSet().getOption("rsize").getResultValue(0));
            SequenceRANDOM rand = new SequenceRANDOM(
                                        Sequence.loadSequences(inFileRandom, Sequence.SequenceType.RANDOM, 0), 
                                        randSize);
            pre.createCutoffs(rand, percentage, basis);
          }
          if(r > 0){
            System.out.println("Read in "+r+" Random Sequences");
          }
            
          pre.createTrx(outFileDat, formatLong);
          IFimProcess fimMax = new FimMAX(userdir+"/fim_maximal", 100, outFileDat, outFileFim);
          Poisson poisson = Poisson.getInstance();
          
          pre.printMotifFrequency();

          poisson.init(pre.getMotifFrequency(),
                       pre.getNonRandomSequenceLength(),
                       pre.getNonRandomNumSequences(),
                       outFileSig,
                       fimMax);

          poisson.CalculatePoisson();
        }
        catch(IOException e) {
          System.out.println("Error reading input or writing output file. Please check paths and names.");
          System.out.println(e.getMessage());
        }
        time = System.currentTimeMillis() - time;
        System.out.println("\nexecution time: "+time/(1000*60)+" min.");
    }
    
    
  /**
   * Returns the Usage string
   */
    public static String getUsage(){
      StringBuffer sb = new StringBuffer();
      sb.append("java -jar mpredict.jar [options] <PWM File in> <Sequences.Fasta in> \n")
        .append("[options]:\n")
        .append("-c <Cutoff file in> Cutoff file format\n")
        .append("-r <Random.Fasta in> .FASTA file format\n")
        .append("-rsize Size of random sequence to create\n")
        .append("\t -b <% cutoff basis>: (p)wm or (s)equence>, default: (p)wm\n")
        .append("\t -per <[0-1]>: .60 = cutoff default \n")
        .append("-l Output results in long format\n")
        ;
      return sb.toString();
    }
  
  
  /**
   * Displays Message, Usage and Exits program 
   * @param message
   */
    public static void doUserExit(String message){
    System.out.println(message);
    System.out.println(getUsage());
    System.exit(1);
  }
  
}
