package mpredict;

import java.io.BufferedReader;
import java.io.FileReader;
import java.io.IOException;

import java.util.ArrayList;
import java.util.List;

public class PWM implements Comparable<PWM> {
    protected ArrayList<float[]> M = new ArrayList<float[]>();
    private int Id=0;
    protected String name="";
    private float pseudo = 0.01f;
    private float minScore=Float.NEGATIVE_INFINITY ;
    private float maxScore=Float.POSITIVE_INFINITY ;
    private static int numPWM=0;
    
    public int getID(){
        return Id;
    }
    
    public PWM() {
    }
    
  /**
   * Load PFM Matrices into Memory and create PWMS
   * @param InFileFWM
   * @param pseudo
   * @return number of PWMS loaded
   * @throws IOException
   */
    public static List<PWM> loadPWMs(String InFileFWM, float pseudo) throws IOException {
      List aPWM = new ArrayList<PWM>();
      BufferedReader br = null;
      int id = 0; //Ids begin here and increment by 1 for each
      
      try{
          PWM pwm=null;
          br = new BufferedReader(new FileReader(InFileFWM));
          String line;            
          // Read in entire file and create PWMs
          while ((line = br.readLine()) != null){
              /*** Creates and reads in PWM till next one detected ***/
              // PWM header
              if(line.startsWith(">")){
                  pwm = new PWM(id++, line, pseudo);
                  aPWM.add(pwm);
              }
              // PWM remaining lines
              else if(pwm != null && !line.isEmpty()) {
                  pwm.setRow(line);
              }
          }
      }
      finally{
          if(br != null)
              br.close();
      }
      return aPWM;
    }    
    
    // Given PFM we create PWM
    protected PWM(int id, String attributes, float pseudo) throws IOException {
        Id = id;
        this.pseudo = pseudo;
        String[] tokens;
        tokens = attributes.split("\t");
        if (tokens.length >= 2 && tokens[0].startsWith(">"))
        {
            String n=tokens[0].split("_")[0].toUpperCase();
            //n = n.replace("V$","");
            this.name = n;
        }
        else {
            throw new IOException("This PFM is not formatted correctly:"+attributes);
        }
        numPWM++;
    }
    
    /**
   * Calculate PWM from PFM
   * Calculate corrected probability of each nucleotide p(b,i) and divide p(b,i) by background probabilities and convert to log-scale.
   * Each call to setRow sets a row of PWM at a time
   * @param values
   * @throws IOException
   */
    protected void setRow(String values) throws IOException{
        String tokens[] = values.split("\\s+");
        if (tokens.length != 4) {
            throw new IOException("PFM not formatted correctly, needs 4 columns {A,C,G,T}:"+values);
        } 
        
        //Background freq of each nucleotide
        double[] bk = new double[4];
        bk[0]=0.290763621993511; //A
        bk[1]=0.202156441211073; //C
        bk[2]=0.205656409776978; //G
        bk[3]=0.301423527018438; //T
        
        float[] row = new float[4];
        float sum=0.0f;            
        for(int i=0; i<4; i++){
            row[i] = Float.parseFloat(tokens[i]);
            sum += row[i];
        }
        
        // Alternate Method
        //double pseudoCount = sum*pseudo;                                    //sum each row + pseudo
        //double total = 4*pseudoCount + sum;
        
        // Alternate Method
        /*
        row[0] = (float)(Math.log( ((row[0] + pseudoCount)/total) / ((bk[0]+bk[3])/2) )); 
        row[1] = (float)(Math.log( ((row[1] + pseudoCount)/total) / ((bk[1]+bk[2])/2) ));
        row[2] = (float)(Math.log( ((row[2] + pseudoCount)/total) / ((bk[1]+bk[2])/2) ));
        row[3] = (float)(Math.log( ((row[3] + pseudoCount)/total) / ((bk[0]+bk[3])/2) ));
        */
        
        // Original Method
        double total = 4*pseudo + sum;
        row[0] = (float)(Math.log( ((row[0] + pseudo)/total) / bk[0]));
        row[1] = (float)(Math.log( ((row[1] + pseudo)/total) / bk[1]));
        row[2] = (float)(Math.log( ((row[2] + pseudo)/total) / bk[2]));
        row[3] = (float)(Math.log( ((row[3] + pseudo)/total) / bk[3]));
        
        M.add(row);
    }
    
    protected float getPosScore(int row, char c){
        c = Character.toUpperCase(c);
        if(row > M.size()){
            System.out.println("Row range "+row+" exceeds Matrix:"+M.size());
            System.exit(1);
        }
        if(c == 'A')
            return M.get(row)[0];
        else if(c == 'C')
            return M.get(row)[1];
        else if(c == 'G')
            return M.get(row)[2];
        else if(c == 'T')
            return M.get(row)[3];
        else return 0; //Non-standard nucleotide
    }
    
    public int getNumRows() {
        return M.size();
    }
    
    public float getMaxScore(){
        if (maxScore == Float.POSITIVE_INFINITY )
            setMaxScore();
        return maxScore;
    }
    
    public float getMinScore(){
        if (minScore == Float.NEGATIVE_INFINITY )
            setMinScore();
        return minScore;
    }
    
    private float setMinScore() {
        minScore=0;
        for (float[] r : M){
            minScore = Min(r)+minScore;
        }
        return minScore;
    }
    
    private float setMaxScore() {
        maxScore=0;
        for(float[] r: M) {
            maxScore = Max(r)+maxScore;
        }
        return maxScore;
    }
    
    private float Min(float[] x){
        float min=x[0];
        for(int i=1; i<x.length; i++) {
            if(x[i]<=min)
                min=x[i];
        }
        return min;
    }
    
    private float Max(float[] y){
        float max=y[0];
        for(int i=1; i<y.length; i++) {
            if(y[i]>=max)
                max=y[i];
        }
        return max;
    }
    
    public int getNumPWM(){
        return numPWM;
    }
        
    public String toString(){
      StringBuffer sb = new StringBuffer();
      sb.append(Id).append(")").append(name).append("\n");
      for(float[] r : M){
          for(int c=0; c<r.length; c++)
              sb.append(r[c]+" ");
          sb.append("\n");
      }
      sb.append("\n");
      return sb.toString();
    }

    public int compareTo(PWM o) {
      return this.name.compareTo(o.name);
    }
    
}
