package mpredict;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.io.InputStreamReader;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.List;

/**
 * This class handles basic command line processing and calling of main objects.
 * MPredict uses predict class which will predict frequent combinations of
 * putative binding sites when given proper parameters.
 * Requires sequences and PFMs for scoring and locating potential motif sites.
 * Erik Ladewig 8/1/2010
 */

public class Poisson {
  private int trxId;
  private float[] motifIDFreq;
  private double[] logC;
  private int seqLen;
  private int numSeq;
  private int avgLen;
  private String fileOut;
  private int[] sigValues;
  double pCRM;
  private static IFimProcess fimProcess;
  
  private static Poisson uniqueInstance = new Poisson();
  
  
  private Poisson(){}
  
  
  public static Poisson getInstance(){
    return uniqueInstance;
  }


  public void init(float[] motifFreq, int seqLength, int numSequences, String fileOut, IFimProcess fimProc) {
    this.trxId = 0;
    this.seqLen = seqLength; // 156672008;
    this.numSeq = numSequences;
    this.fileOut = fileOut;
    this.motifIDFreq = motifFreq;
    this.fimProcess = fimProc;
  }
  
  
  /**
   * Runs fim process that feeds results directly to calculate significance of all entries
   * @throws IOException
   */
  public void CalculatePoisson() throws IOException{
    int m = 10000<this.numSeq ? 10000 : this.numSeq;
    
    logC = new double[m];
    for(int i=0; i<m; i++){
      logC[i] = logc(numSeq, i);
    }
    
    if(numSeq <= 0){
      throw new IOException("Can not calculate for 0 sequences");
    }
    
    this.avgLen = (int)(seqLen/numSeq);//231790); // average non coding length
    
    File f = new File(fimProcess.getOutFile());
    
    BufferedWriter bw = null;
    BufferedWriter bwFim = null;
    int lineCount=0;
    try{
      final Process p = fimProcess.getProcess();
      BufferedReader input = new BufferedReader(new InputStreamReader(p.getInputStream()));
      BufferedReader stdError = new BufferedReader(new InputStreamReader(p.getErrorStream()));
      bw = new BufferedWriter(new FileWriter(fileOut, false));
      if(f != null){
        bwFim = new BufferedWriter(new FileWriter(fimProcess.getOutFile(), false));
      }
      
      // Our process outputs to stdout and feeds the significance calculation method
      String line="";
      while( (line = input.readLine())!= null) {
        // In case user wants fim output as well
        if(bwFim != null){
          bwFim.write(line+"\n");
        }
        
        if(CalculateSignificance(line)){
          bw.write(this.toString());
          lineCount++;
        }
      }
      
      int exitVal = p.waitFor(); // wait for our fim process to finish
      if(exitVal != 0){
        System.out.println("Fim encountered errors:"+input.readLine());
        while( (line = stdError.readLine()) != null){
          System.out.println(line);
        }
      }
    }
    catch(InterruptedException iie){
      iie.printStackTrace();
    }
    finally{
        if( bw != null)
          bw.close();
        if( bwFim != null)
          bwFim.close();
    }    
    System.out.println("Significance scoring generated "+lineCount+" lines to "+fileOut);
  }
  
  
  /**
   * Calculates significance of FIM Result
   * @param fimResult
   * @return String to write
   */
  private boolean CalculateSignificance(String fimResult){
    String[] tokens = fimResult.trim().split("\\s+");
    int n = tokens.length-1;
    int[] motifs = new int[n];
    sigValues = new int[0];
    int support;    
    trxId++;
    
    //Support is not found or less than # motifs
    if(n < 2)
      return false;
    
    try{
      for(int i=0; i<motifs.length; i++){
        motifs[i] = Integer.parseInt(tokens[i]);
      }
      support = Integer.parseInt( tokens[n].replaceAll("\\D+", "") );
    }
    catch (NumberFormatException ne){
      System.out.println("NAN "+tokens[n]);
      return false;
    }

    // Probability
    double pc = 0;
    double lambda = 0;
    for(int i=0; i<n; i++){
      lambda = Math.log(motifIDFreq[motifs[i]]);
      for(int j=0; j<n; j++){
        if(j!=i){
          double interim = Math.log(1-Math.exp(-avgLen*motifIDFreq[motifs[j]]));
          lambda+=interim;
        }
      }
      pc+=Math.exp(lambda);
    }
    
    pc = pc*avgLen*2;
    
    double logpc = Math.log(pc);
    double logpcn = Math.log(1-pc);
    
    //if logpcm will be undefined
    if(Double.isNaN(logpcn))
      return false;
    
    pCRM = 1.0;
    for(int j=0; j<support; j++){
      double sigma = (double)Math.exp(logC[j]+j*logpc+(numSeq-j)*logpcn);
      pCRM = pCRM - sigma;
    }
    
    // Correction Method
    double sumcm = 0.0;
    for(int i=2; i<=n; i++){
      sumcm += Math.exp(logc(motifIDFreq.length-1, i));
    }
    pCRM= pCRM*sumcm;
    
    if(pCRM >= 0.05){
      return false;
    }
    
    sigValues = Arrays.copyOf(motifs, motifs.length+1);
    sigValues[motifs.length] = support;
    
    return true;
  }
    
  
  /**
   * Computes Log Value of combination
   * @param total
   * @param sample
   * @return
   */
  private double logc(int total, int sample){
    if(sample == 0)
      return 0;

    double z=0;
    if(sample > 0){
      for(int k=1; k<=sample; k++){
        z+=(Math.log((double)(total-k+1))-Math.log((double)k));
      }
    }
    return z;
  }
  
  
  public String toString(){
    String ret = trxId+"\t"+(sigValues.length-1);
    for(int i=0; i<sigValues.length; i++) {
      ret+="\t"+sigValues[i];
    }
    ret+="\t"+pCRM+"\n";
    return ret;
  }  
}
