package mpredict;


import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.text.DecimalFormat;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.TreeMap;
import java.util.TreeSet;

/**
 * This class combines objects together for processing and gathering of data.
 * MPredict with predict frequent combinations of putative binding sites when given parameters.
 * Requires sequences and pwms for scoring and locating potential motif sites.
 * Erik Ladewig 8/1/2010
 */

public class Predict {
    
   /** Class variables **/
    private List<PWM> aPWM;
    private float[] cutOffs;
    private List<Sequence> aSeqs;
    private float percent = 0.f;
    private boolean Appendtrx = false;
    private int[] Moccur;
    private int nonRandSeqLength = 0;
    
    public enum CutoffBasis{
      PWM('p'),
      
      SEQUENCE('s');
      
      private char c;
      private CutoffBasis(char c){
        this.c = c;
      }
    };
    
    public void setPercentCut(float f){
      percent = f;
    }
    
    public float[] getMotifFrequency(){
      if(Moccur == null)
        return new float[]{};
      
      getNonRandomSequenceLength();
      if(nonRandSeqLength==0)
        return new float[] {};
      
      float[] motifIDFreq = new float[Moccur.length+1];
      for(int i=0; i<Moccur.length; i++){
        // Translation: moccur stores IDs starting at 0, however .dat file stores them starting at 1
        motifIDFreq[i+1] = (float)Moccur[i]/nonRandSeqLength; //156672008; 
      }      
      return motifIDFreq;
    }
    
    public int getNonRandomSequenceLength(){
      if(nonRandSeqLength == 0){
        for(Sequence s : aSeqs){
          nonRandSeqLength += s.sequence.length();
        }
      }
      return nonRandSeqLength;
    }
    
    public int getNonRandomNumSequences(){
      return aSeqs.size();
    }
    
  /**
   * Load PFM Matrices into Memory and create PWMS
   * @param InFileFWM
   * @param pseudo
   * @return number of PWMS loaded
   * @throws IOException
   */
    public int loadPWMs(String InFileFWM, float pseudo) throws IOException {
      aPWM = PWM.loadPWMs(InFileFWM, pseudo);
      return aPWM.size();
    }
    
    
  /**
   * Load cutoff file into Memory
   * @param InFileC
   * @return number of cutoffs loaded, should match number of PWMS.
   * @throws IOException
   */
    public int loadMotifCutOffs(String InFileC) throws IOException{
      if(InFileC == null)
        return 0;
    
      cutOffs = new float[aPWM.size()];
      BufferedReader br = null;
      int i=0;
      try{
          br = new BufferedReader(new FileReader(InFileC));
          String line;

          while( (line=br.readLine()) != null){
              if(line.startsWith(">")) {
                  String[] split = line.split("\\s+");
                  if(split.length >=3){
                      try{
                          float val = Float.parseFloat(split[1]);
                          cutOffs[i] = val;
                          i++;
                      }
                      catch(NumberFormatException f) {
                          System.out.println("Cutoff in incorrect format!");
                          break;                            
                      }
                  }
                  else{
                      System.out.println("Cutoff in incorrect format!");
                      break;
                  }
              }
          }
      }
      finally{
          if(br != null)
              br.close();
      }
      if(i!=aPWM.size())
          throw new IOException("Cutoff file number of entries not equal to # PWMs "+i+"!="+aPWM.size());
      return cutOffs.length;
    }
  
  
  /**
   * Load sequences into Memory using File format .fasta, example:
   * >mm8 chr1:5073158_5073258
   * TCAGCTGCGGTCACGTGCGGCTGCTCACGTGCCAGACGCTGCTGTCACTT
   * @param InFileTF
   * @return
   * @throws IOException
   */
    public int loadSequences(String InFileTF, Sequence.SequenceType type) throws IOException{
      aSeqs = new ArrayList( Sequence.loadSequences(InFileTF, type, 0) );
      return aSeqs.size();
    }
   
   
    /**
    * Creates score cutoffs from distribution bins
    * @param s
    */
    protected void createCutoffs(Sequence s, float percentile, CutoffBasis basis){
     System.out.println("Generating cutoff scores from sequence of size "+s.sequence.length()+"...");
     if(aPWM.size() == 0)
       return;
     
     final int NUMBINS = 100000;
     cutOffs = new float[aPWM.size()];
    
     System.out.println("Motif\tCurrentScore\t%");
     for(PWM p : aPWM){
       float t = p.getMinScore();
       List<MotifScore> passed = s.scoreSequence(p, t);
       //Base min and max score on range of sequence or PWM(default).
       float maxScore = p.getMaxScore();
       float minScore = p.getMinScore();
       if(basis == CutoffBasis.SEQUENCE){
         maxScore = s.getMaxScore();
         minScore = s.getMinScore();
       }
       float[] bins = new float[NUMBINS];
       int maxBin = NUMBINS-1;
       float step = (float) (maxScore - minScore) / (NUMBINS-1);
    
       // Non-uniform distribution count scores and place in bins
       for(MotifScore m : passed){
           int bin = (int)Math.floor((m.score-minScore) / step);
           try{ bins[bin] += 1; } //Put in bin
           catch (ArrayIndexOutOfBoundsException e ){
               System.out.println(e.getLocalizedMessage()+" @ bin "+bin+" for score:"+m.score+" min:"+minScore+" max:"+maxScore+" step:"+step);
               System.exit(1);
           }
       }
    
       int sum = 0;
       // Convert bin contents to percentage
       for(int i=0; i< bins.length; i++){
           sum += bins[i];
           bins[i] = (float) sum/passed.size();
       }
    
       // Sets minimum score to be in percentile
       int k=0;
       while(k<maxBin){
           if(bins[k]<=percentile)
               k++;
           else{
               if(k>0)
                   k=k-1;
               break;
           }
       }
       float score = (float)(k)*step+minScore;
       cutOffs[p.getID()] = score;
       
       if(basis == CutoffBasis.SEQUENCE)
        s.resetScores();
      
       System.out.println(p.name+"\t"+String.format("%2.6f", score)+"\t"+String.format("%2.8f",bins[k]));
     }
    }    
    
    
   /**
   * Records scores to file
   * @param OutDatFile
   * @throws IOException
   */
    public void createTrx(String OutDatFile, boolean formatLong) throws IOException{
      System.out.print("Creating transactions process...");
      BufferedWriter bw = null;
      try{
        bw = new BufferedWriter(new FileWriter(OutDatFile, Appendtrx));
        for(Sequence s : aSeqs){
          List<MotifScore> aPass = recordScores(s);
          List<MotifScore> pass = createTRX(aPass);

          if(formatLong)
            writeTrxOutput(s, percent, pass, bw);
          else
            writeDatOutput(pass, bw);

          calculate_frequency(pass);
        }
      }
      finally{
          if( bw != null)
              bw.close();
      }
      System.out.println("done.");
    }
    
    
  /**
   * Gathers sequences with scores within criteria
   * @param s
   * @return
   * @throws IOException
   */
    private List<MotifScore> recordScores(Sequence s) throws IOException{
      List<MotifScore> aPass = new ArrayList<MotifScore>(aPWM.size());
      for(int i=0; i<aPWM.size(); i++){                  // For sequence s get a scores for each PWM
          PWM p = aPWM.get(i);
          float t = cutOffs[p.getID()];                  // Get defined cutoff
          List<MotifScore> passed = s.scoreSequence(p, t);  // Get the scored sequence segments using pwm and cutoff
          aPass.addAll(passed);
      }
      return aPass;
    }
    
    
   /**
   * Create Transaction for output
   * @param aPass
   * @return
   * @throws IOException
   */
    private List<MotifScore> createTRX(List<MotifScore> aPass) throws IOException{
      if(aPass.size() == 0){
          return new ArrayList<MotifScore>();
      }
      
      int overlapDistance = 3;
      Collections.sort(aPass, new MotifScore.byMotifScoreDesc());    // Sort by score
      List<MotifScore> keep = new ArrayList<MotifScore>();
      
      // Remove overlapping motifs
      MotifScore remove = new MotifScore(null, null, Sequence.SequenceStrand.Plus,-1,-1);
      for(int i=0; i<aPass.size()-1; i++){
          if( aPass.get(i) != remove){
              MotifScore top = aPass.get(i);
              for(int j=i+1; j<aPass.size(); j++){
                  MotifScore n = aPass.get(j);
                  if(n != remove)
                      if(Math.abs(n.offset - top.offset) <= overlapDistance )
                          aPass.set(j, remove);
              }
          }
      } 
      for(MotifScore k : aPass){
          if(k!=remove)
              keep.add(k);
      } // End remove
      
      return keep;
    }
 
  
    /**
    * Stores Motif occurrence rate
    * @param pass
    */
    private void calculate_frequency(List<MotifScore> pass){
       //Initialize motif occurrence map
      if(Moccur == null){
       Moccur = new int[aPWM.size()];
      }
       
       for( MotifScore m : pass ){
         int pId = m.motif.getID();
         Moccur[pId]++;
       }
     }
      
   
    /**
    * Output for each motif its name and the number of times it occurs
    */
    public void printMotifFrequency(){
      float[] motiffreq = getMotifFrequency();
      if(motiffreq.length == 0)
        return;
      
      DecimalFormat formatter = new DecimalFormat("0.######E0");
      System.out.println("Motif\tFrequency");
      for(int i=0; i<motiffreq.length-1; i++){
       System.out.println(aPWM.get(i).name+"\t"+formatter.format((Float)motiffreq[i+1]));
      }
     }  

    
   /**
   * Writes transactions to outFile using TRX format.
   * @param s
   * @param percent
   * @param pass
   * @param OutFile
   * @throws IOException
   */
    private final void writeTrxOutput(Sequence s, float percent, List<MotifScore> pass, BufferedWriter bw) throws IOException{
        StringBuilder sb = new StringBuilder();
        
        if(pass != null && !pass.isEmpty() ){
            Collections.sort(pass);
        }
        for(MotifScore m : pass){
          sb.append((m.motif.getID()+1)+"\t"+m.strand+"\t"+m.offset+"\t"+String.format("%6.4f", m.score)+"\t");
        }
        sb.append(pass.size());
        bw.write(s+" "+percent+"\n");
        bw.write(sb.toString());
        bw.write("\n");
    }
    
    
   /**
   * Writes dat to outFile using DAT format
   * @param s
   * @param pass
   * @param OutFile
   * @throws IOException
   */
    private final void writeDatOutput(List<MotifScore> pass, BufferedWriter bw) throws IOException{
      StringBuilder sb = new StringBuilder();
      int[] uniques = new int[aPWM.size()+1];
      if(!pass.isEmpty()){
        //Get unique entries
        Collections.sort(pass);
      }
      for(MotifScore m : pass){
        // reformat trx output so numbers start at 1 not 0
        uniques[m.motif.getID() + 1] = 1;
      }
      // This is so we print unique entries which exist
      for(int i=0; i<uniques.length; i++){
        if(uniques[i] == 1)
          sb.append(i).append(" ");
      }
      sb.append(" ");
      bw.write(sb.toString());
      bw.write("\n");
    }    
     
    
   /**
   * Writes PWM to outFile
   * @param outFile
   * @throws IOException
   */
    public void writePWM(String outFile) throws IOException {
        if (aPWM == null){
            System.out.println("Matrix DNE!");
            return;
        }
        
        for(PWM m : aPWM)
        {
            System.out.println(m);// "min score:"+m.getMinScore()+" max score:"+m.getMaxScore());
        }
        
        BufferedWriter outStream=null;
        try{
            outStream = new BufferedWriter(new FileWriter(outFile, false));
            for(PWM m : aPWM)
            {
                outStream.write(m.toString()); // "min score:"+m.getMinScore()+" max score:"+m.getMaxScore());
            }
        }
        finally{
            if (outStream != null)
                outStream.close();
        }          
    }     
   
   
  /**
   * Output Cutoff Scores
   */
    public void outputCutoffs(){
     DecimalFormat df = new DecimalFormat("00.00");
     for(int i=0; i<cutOffs.length; i++){
       System.out.println(aPWM.get(i).name+"\t"+df.format(cutOffs[i]));
     }
   }
  
    
    public void outputSequences(){
      for(Sequence s: aSeqs){
        s.printTF();
      }
    }

}
