package mpredict;

import java.io.BufferedReader;
import java.io.FileReader;
import java.io.IOException;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.List;
import java.util.Random;

public abstract class Sequence {
  String header;
  String sequence= "";
  float maxScore= Float.NEGATIVE_INFINITY;
  float minScore= Float.POSITIVE_INFINITY;
  public enum SequenceType{
    STANDARD('s'),
    
    RANDOM('r');
    
    private char c;
    private SequenceType(char c){
      this.c = c;
    }
  };
  
  public enum SequenceStrand{
      Plus('+'),
      Minus('-');
      
      final char code;
      
      SequenceStrand(char code){
          this.code = code;
      }
      
      public String toString(){
          return code+"";
      }
  }  
  
  protected abstract void setAttributes(String attributes) throws IOException, NumberFormatException;
  
  private void appendSequence(String segment){
    this.sequence = this.sequence + segment;//.trim();
  }
  
  public static Sequence create(SequenceType type, String attributes)throws IOException{
      if(type == SequenceType.STANDARD )
          return new SequenceFASTA(attributes);
      else if(type == SequenceType.RANDOM )
          return new SequenceRANDOM(">Random Background");
      return null;
  }  
  
/**
 * Load sequences into Memory using File format .fasta, example:
 * >mm8 chr1:5073158_5073258
 * TCAGCTGCGGTCACGTGCGGCTGCTCACGTGCCAGACGCTGCTGTCACTT
 * @param InFileTF
 * @param enough
 * @return
 * @throws IOException
 */
  public static Collection<Sequence> loadSequences(String InFileTF, SequenceType type, int enough) throws IOException {
    List<Sequence> aSequences = new ArrayList<Sequence>();
    BufferedReader br = null;
    Sequence s = null;
    if(type == SequenceType.STANDARD)
      s = new SequenceFASTA();
    else if (type == SequenceType.RANDOM)
      s = new SequenceRANDOM();
    else
      throw new IOException("Unknown Sequence type:"+type);
    
    try{
        br = new BufferedReader(new FileReader(InFileTF));
        String line;
        
        while( (line=br.readLine()) != null ){
            if (line.startsWith(">"))
            {
                s = new SequenceFASTA(line);
                aSequences.add(s);
            }
            else if(s != null){
                s.appendSequence(line);
            }
            else{
                throw new IOException("This fasta sequence does not have a header beginning with >");
            }
            if(enough > 0 && aSequences.size()>= enough)
                break;
        }
    }
    finally{
        if (br != null)
            br.close();
    }
    return aSequences;
  }    
  
  protected boolean isStandardNucleotide(char c){
      c = Character.toUpperCase(c);
      if(c=='A' || c=='C' || c=='G' || c== 'T')
          return true;
      else
          return false;
  }     
  
  protected String getCDNA(){//ReverseComplementaryChain
    StringBuilder sb = new StringBuilder(sequence).reverse();
    for(int i=0; i<sb.length(); i++ ) {
        switch(sb.charAt(i))                
        {
        case 'A'    : sb.setCharAt(i,'T');break;
        case 'C'    : sb.setCharAt(i,'G');break;
        case 'G'    : sb.setCharAt(i,'C');break;
        case 'T'    : sb.setCharAt(i,'A');break;
        default     : break;
        }
    }
    return sb.toString();
  }
  
  protected void resetScores(){
    maxScore = Float.NEGATIVE_INFINITY;
    minScore = Float.POSITIVE_INFINITY;
  }
  
  protected float getMinScore(){
      return minScore;
  }  
  
  protected float getMaxScore(){
      return maxScore;
  } 
  
  protected void setMinScore(float score){
      if(score < minScore)
          minScore = score;
  }  
  
  protected void setMaxScore(float score){
      if(score > maxScore) {
          maxScore = score;
      }
  }      
    
  protected void printTF(){
      System.out.println(header);
      int n=20000;
      int wrap = ((Double)(Math.floor(sequence.length()/n))).intValue();
      int i=0;
      if( sequence.length() > n )
      {
          for(i=0; i<wrap; i++ )
          {
              System.out.println(sequence.substring(i*n, (i+1)*n ));
          }
          int diff = sequence.length()-(i*n);
          if(diff != 0)
              System.out.println(sequence.substring((i*n)-1, i*n+diff-1));
      }
      else
          System.out.println(sequence);
  }  
  
  protected StringBuffer tfOutput(){
      StringBuffer sb = new StringBuffer(header);
      int wrap = sequence.length();
      int n=50;
      if (sequence.length() > n)
      {
          wrap = ((Double)(Math.floor(sequence.length()/n))).intValue();
          int i=0;
          for(i=0; i<wrap; i++ )
          {
              sb.append(sequence.substring(i*n, (i+1)*n )).append("\n");
          }
          int diff = sequence.length()-(i*n);
          if(diff != 0)
          {
              sb.append(sequence.substring((i*n)-1, i*n+diff-1)).append("\n");
          }
      }
      else
          sb.append(sequence);
      return sb;
  }   
  
  public String toString(){
    return header;
  }
  
  /**
   * Records each sequences score against a PWM and a threshold
   * @param s
   * @param p
   * @param threshold
   * @return
   */
  protected final List<MotifScore> scoreSequence(PWM p, float threshold){
      List<MotifScore> aMpass = new ArrayList<MotifScore>();
      String cDNA = this.getCDNA();
      int i=0, n=0;        
      int L=this.sequence.length();
      int W=p.getNumRows();
      float score=0, rscore=0;
      
      for( int j=0; j<L; j++){
        i=0;        // i will traverse motif length
        n=j;        // n starts @ j and traverses sequence
        score=0;    
        rscore=0;
        while (n < L && i < W ){
            if(!this.isStandardNucleotide(this.sequence.charAt(n)))
                break;
            score += p.getPosScore(i, this.sequence.charAt(n));
            rscore += p.getPosScore(i, cDNA.charAt(n));
            i=i+1;
            n=n+1;               
            if(i == W)
            {                             
                if (score >= threshold){
                    MotifScore pass = new MotifScore(p, this, SequenceStrand.Plus , j, score);
                    aMpass.add(pass);
                    this.setMaxScore(score); this.setMinScore(score);
                }
                if (rscore >= threshold){
                    MotifScore rpass = new MotifScore(p, this, SequenceStrand.Minus, L-n, rscore);
                    aMpass.add(rpass);
                    this.setMaxScore(rscore); this.setMinScore(rscore);
                }
            }
        }
      }
      return aMpass;
  }      
  
  
  private static StringBuilder shuffle(StringBuilder sb){
      for(int i=0; i<sb.length(); i++){
          char c = sb.charAt(i);
          int pick = new Random().nextInt(sb.length());
          char r = sb.charAt(pick);
          sb.setCharAt(i, r);
          sb.setCharAt(pick, c);
      }
      return sb;
  }
  
}
